""" Namespaces for testing
Yes, should go in a test packages.
Here for now to test a package importing from itself
"""


import polars as pl

class CustomNamespace:
    def __init__(self, ldf: pl.LazyFrame):
        self._ldf = ldf

    def custom_method(self, x: int) -> pl.LazyFrame:
        """my docstring"""
        return self._ldf

class CustomNamespaceByPath:
    def __init__(self, ldf: pl.LazyFrame):
        self._ldf = ldf

    def custom_method(self, x: int) -> pl.LazyFrame:
        """my docstring"""
        return self._ldf


class EntryPointNameSpace:
    def __init__(self, ldf: pl.LazyFrame):
        self._ldf = ldf

    def custom_method(self, x: int) -> pl.LazyFrame:
        """my docstring"""
        return self._ldf






import importlib
from typing import Callable, Dict, Optional, Type, Union
from polugins._types import ExtensionClass
import polars as pl
import sys


# TODO: Stolen from pluggy, remember license and stuff
if sys.version_info >= (3, 8):
    from importlib import metadata as importlib_metadata
else:
    # TODO: remember that this is only needed for python < 3.8
    import importlib_metadata  # ignore


def get_entrypoints():
    entry_points = {
        ExtensionClass.EXPR: [],
        ExtensionClass.SERIES: [],
        ExtensionClass.LAZYFRAME: [],
        ExtensionClass.DATAFRAME: [],
    }
    for dist in list(importlib_metadata.distributions()):
        for ep in dist.entry_points:
            if ep.group.startswith("polugins"):
                entry_points[ExtensionClass.from_entrypoint_group(ep.group)].append(ep)
    return entry_points

def load_setuptools_entrypoints():
    """
    TODO: Stolen from pluggy, remember license and stuff
    """
    all_entry_points = get_entrypoints()
    namespaces = {
        ExtensionClass.EXPR: {},
        ExtensionClass.SERIES: {},
        ExtensionClass.LAZYFRAME: {},
        ExtensionClass.DATAFRAME: {},
    }
    for namespace_type, entry_points in all_entry_points.items():
        for ep in entry_points:
            if ep.group.startswith("polugins"):
                namespace = ep.load()
                namespaces[namespace_type][ep.name] = namespace
    return namespaces

NamespaceDict = Dict[str, Union[Type, str]]

def register_namespaces(
    lazyframe_namespaces: NamespaceDict = {},
    dataframe_namespaces: NamespaceDict = {},
    expression_namespaces: NamespaceDict = {},
    series_namespaces: NamespaceDict = {},
    entrypoints: bool = True,
):
    for name, namespace in lazyframe_namespaces.items():
        ExtensionClass.LAZYFRAME.register(name, namespace)
    for name, namespace in dataframe_namespaces.items():
        ExtensionClass.DATAFRAME.register(name, namespace)
    for name, namespace in expression_namespaces.items():
        ExtensionClass.EXPR.register(name, namespace)
    for name, namespace in series_namespaces.items():
        ExtensionClass.SERIES.register(name, namespace)
    if entrypoints:
        ep_namespaces = load_setuptools_entrypoints()
        for extension_class, namespaces in ep_namespaces.items():
            for name, namespace in namespaces.items():
                extension_class.register(name, namespace)



from enum import Enum
import importlib
from pathlib import Path
from typing import Callable, Type, Union
from typing_extensions import Self, TypeVar
import polars as pl

NS = TypeVar("NS")

class ExtensionClass(Enum):
    EXPR = "expr"
    SERIES = "series"
    LAZYFRAME = "lazyframe"
    DATAFRAME = "dataframe"

    def to_reg_function(self) -> Callable:
        if self == ExtensionClass.EXPR:
            return pl.api.register_expr_namespace
        if self == ExtensionClass.SERIES:
            return pl.api.register_series_namespace
        if self == ExtensionClass.LAZYFRAME:
            return pl.api.register_lazyframe_namespace
        if self == ExtensionClass.DATAFRAME:
            return pl.api.register_dataframe_namespace
        raise TypeError  # will never happen

    def register(self, name: str, namespace: Union[str, Type[NS]]) -> Type[NS]:
        if type(namespace) == str:
            module, object = namespace.split(":")
            namespace_module = importlib.import_module(module)
            namespace = getattr(namespace_module, object)
        if self == ExtensionClass.EXPR:
            return pl.api.register_expr_namespace(name)(namespace) # type: ignore
        if self == ExtensionClass.SERIES:
            return pl.api.register_series_namespace(name)(namespace) # type: ignore
        if self == ExtensionClass.LAZYFRAME:
            return pl.api.register_lazyframe_namespace(name)(namespace) # type: ignore
        if self == ExtensionClass.DATAFRAME:
            return pl.api.register_dataframe_namespace(name)(namespace) # type: ignore
        raise TypeError  # will never happen, poor mans pattern matching.

    @property
    def import_path(self) -> Path:
        if self == ExtensionClass.EXPR:
            return Path("polars", "expr", "expr")
        if self == ExtensionClass.SERIES:
            return Path("polars", "series", "series")
        if self == ExtensionClass.LAZYFRAME:
            return Path("polars", "lazyframe", "frame")
        if self == ExtensionClass.DATAFRAME:
            return Path("polars", "dataframe", "frame")
        raise TypeError  # will never happen


    @classmethod
    def from_entrypoint_group(cls, group: str) -> Self:
        namespace_type = group.split(".")[1]
        return cls(namespace_type)




from pathlib import Path
from mypy.stubgen import Options as StubOptions, generate_stubs
import sys

from polugins.main import get_entrypoints
from polugins._types import ExtensionClass
import ast


def generate_polars_stub(output_dir: Path):
    modules = [".".join(enum.import_path.parts) for enum in ExtensionClass]
    options = StubOptions(
        # 'module_file' contains the base class
        #  onto which dynamic methods are registered
        files=[],
        output_dir=str(output_dir),
        include_private=True,
        export_less=False,
        # standard params (not auto-defaulted)
        pyversion=sys.version_info[:2],
        interpreter=sys.executable,
        ignore_errors=False,
        parse_only=False,
        no_import=False,
        search_path=["."],
        doc_dir="",
        packages=[],
        modules=modules,
        verbose=True,  # TODO: change this, but nice for debugging now
        quiet=False,
    )
    generate_stubs(options)


def main():
    output_dir = Path("typings")

    # TODO: only generate stubs for extensions class that are
    # actually extended
    generate_polars_stub(output_dir)

    entry_points = get_entrypoints()

    for extension_class, entrypoints in entry_points.items():
        if entrypoints:
            stub_path = (output_dir / extension_class.import_path).with_suffix(".pyi")
            stub_ast = ast.parse(stub_path.read_text(), type_comments=True)
            new_class_nodes = []
            modules_to_import = set()
            for node in stub_ast.body:
                if (
                    isinstance(node, ast.ClassDef)
                    and node.name.casefold() == extension_class.value.casefold()
                ):
                    for ep in entrypoints:
                        annotation_name = ep.value.replace(":", ".")
                        new_node = ast.AnnAssign(
                            target=ast.Name(id=ep.name),
                            annotation=ast.Name(id=annotation_name),
                            simple=1,
                        )  # no idea what `simple` does, but it breaks without it
                        new_class_nodes.append(new_node)
                        modules_to_import.add(annotation_name.split(".")[0])
                    node.body.extend(new_class_nodes)
            for name in modules_to_import:
                # prepend to have imports first
                stub_ast.body.insert(0, ast.Import(names=[ast.Name(id=name)]))
            stub_path.write_text(ast.unparse(stub_ast))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['polugins']

package_data = \
{'': ['*']}

install_requires = \
['importlib-metadata>=6.6.0,<7.0.0',
 'mypy>=1.3.0,<2.0.0',
 'polars>=0.18.0,<0.19.0']

setup_kwargs = {
    'name': 'polugins',
    'version': '0.2.0',
    'description': 'Plugin system for Polars.',
    'long_description': '# Polugins\n\nEarly PoC for a "plugin" system for polars.\n\nNot production ready - barely even alpha ready. Only uploading it now for discussion and to hog the genius package name.\n\nIt\'s meant to solve two issues with using polars API extensions:\n\n- You need to import the namespace to trigger the registration, even if you dont need anything from the namespace module.\n\n- Extensions breaks static typing.\n\nThe idea is to describe a single way to expose and collect API extensions - especially for third party packages - \nand then used this discoverbility to also generate type stubs with the added typing from the extensions.\n\nUsers can either call `register_namespaces` themselves or import polars through `polugins.polars` instead.\nLint rules can then be used to enforce that nothing is imported from polars outside of these locations.\n\nThis is still a bit annoying no matter what, unless polars does the import natively.\n\n## Package example\n\nPackages can expose namespace through entry points called `polugins.<class>`, forexample `polugins.lazyframe`.\n\nIf building with poetry you should add this to your `pyproject.toml`:\n\n```toml\n[tool.poetry.plugins."polugins.<class>"]\n"<accessor_name>" = "<path.to.module:NameSpace>"\n\n# Concrete example:\n\n[tool.poetry.plugins."polugins.lazyframe"]\n"external" = "example_package:PackageNamespace"\n```\n\nSee `tests/pkgs/example_package` for a example.\n\n## Usage example\n\nNamespaces can be registered in three ways:\n\n- By module path\n- As imported module\n- From entry points\n\nSay that you have a package `my_package` with two `LazyFrame` namespaces - `MyNamespace` and `AlsoMyNamespace` and you use an\nexternal package `example-package` that exposes a `LazyFrame` namespace called `external`.\nAfter installing it, namespaces can be registered like so:\n\n```python\nfrom polugins import register_namespaces\nimport polars as pl\nfrom my_package import MyNamespace\n\nregister_namespaces(\n    lazyframe_namespaces={\n        \'my_namespace\': MyNamespace,\n        \'also_my_namespace\': "my_package:AlsoMyNamespace" # Note the `:` to separate module path from object\n    },\n    entrypoints=True # Loads from example-package\n  )\n\n# All namespaces are now registered\n(\n  pl.LazyFrame()\n  .external.some_method()\n  .my_namespace.some_method()\n  .also_my_namespace.some_method()\n)\n```\n\nYou need to make sure that you have called `register_namespaces` before trying to use any of those namespaces.\n\nAs an alternative, polars is re-exported through `polugins` such that entrypoint namespaces are automagically registered:\n\n```python\nfrom polugins import pl\n\npl.LazyFrame().external.some_method(x=1)\n```\n\nNote that this only registers entrypoint namespaces (for now).\n\n### Generate types\n\nRun `polugins.create_types.py` to create type stubs. Will create the directory `.typings`.\n\nOnly works for entrypoint namespaces. You can add your own namespaces to the types if you want.\nIt\'s fairly straight forward, just import the namespace on top of the `pyi` and then add an annotation\nto the class with the namespace.\n\nWill be a CLI tool at some point.\n\n## Implementation\n\nJust a thin wrapper around `polars.api.register_x_namespace` and then using `importlib.metadata` to collect\nnamespaces from external packages.\n\nTypes are generated by using mypy to create stubs for lazyframe, dataframe, expr and series and then adding the\nnamespaces to these type stubs.\n\n## Notes\n\nIt\'s still not entirely clear how an application should register its own namespaces.\nEntry points can be used but\n\n- (1) an application might just want to use its own namespaces and not expose them and \n- (2) its a bit annoying, because changes to entrypoints are only registered when the package is installed, even in editable mode (I think).\n\n\n\n',
    'author': 'StefanBRas',
    'author_email': 'opensource@bruhn.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
